
#include "UtilTest.hpp"

#include <cstdio>
#include <cstdlib>
#include <stdexcept>
#include <cppunit/config/SourcePrefix.h>
#include <cppunit/TestAssert.h>
#include "fixerutils.hpp"

using namespace std;

CPPUNIT_TEST_SUITE_REGISTRATION(UtilTest);

#define BSIZE(b) ((int) (sizeof b / sizeof b[0]))
#define BBYTES(b, delta) (sizeof b[0] * (BSIZE(b) + (delta)))

const int UtilTest::b[] = {0, 1, 2, 3, 4, 5, 6, 7};

/**
	Returns some arbitrary value in the specified quantization step.
*/
float UtilTest::randomValueForStep(int step, int n) {
	/*
		.999 * realLowerBound * n = step -> realLowerBound = step / (.999 * n).
		
		x > 1 -> 1/((x-1)/x) = x/(x-1) < (x+1)/(x-1).
		
		x := 1000 -> realLowerBound < (step * 1001.0f) / (999.0f * n).
	*/
	float lowerBound = (step * 1001.0f) / (999.0f * n);
	
	if ((int) (0.999f * (lowerBound * n)) != step)
		throw runtime_error("Couldn't find lower bound - too many steps!");
	
	/*
		.999 * realUpperBound * n = step+1 -> realUpperBound = (step+1) / (.999 * n).
		
		x > 1 -> 1/((x-1)/x) = x/(x-1) > x/x = 1.
		
		x := 1000 -> realUpperBound < (float) (step+1) / n.
	*/
	float upperBound = (float) (step+1) / n;
	
	if ((int) (0.999f * (upperBound * n)) != step)
		throw runtime_error("Couldn't find upper bound - too many steps!");
	
	return lowerBound + (upperBound - lowerBound)*((rand() % 10000) / 10000.0f);
}

void UtilTest::assertRequant(
	int nBefore, int nRemoved, const int *removed, int nInserted, const int *inserted,
	int nValues, const int *beforeValues, const int *afterValues)
{
	char beforeString[128];
	int nAfter = nBefore - nRemoved + nInserted;
	
	for (int i = 0; i < nValues; i++) {
		float randomizedValue = randomValueForStep(beforeValues[i], nBefore);
		int actualOutput = getStep(
			requantize(nBefore, randomizedValue, nRemoved, removed, nInserted, inserted),
			nAfter);
		
		sprintf(beforeString, "Input value: %d", beforeValues[i]);
		CPPUNIT_ASSERT_EQUAL_MESSAGE(beforeString, afterValues[i], actualOutput);
	}
}


void UtilTest::setUp() {
	srand(time(NULL));
}

void UtilTest::tearDown() {}


void UtilTest::testDoNothing() {
	assertRequant(BSIZE(b), 0, NULL, 0, NULL, BSIZE(b), b, b);
}


void UtilTest::testRemoveFirst() {
	int r = 0; // remove 0.
	int a[] = {0, 0, 1, 2, 3, 4, 5, 6}; // b remove 0.
	assertRequant(BSIZE(b), 1, &r, 0, NULL, BSIZE(b), b, a);
}

void UtilTest::testRemoveLast() {
	int r = 7; // remove 7.
	int a[] = {0, 1, 2, 3, 4, 5, 6, 6}; // b remove 7.
	assertRequant(BSIZE(b), 1, &r, 0, NULL, BSIZE(b), b, a);
}

void UtilTest::testRemoveN() {
	int r[] = {1, 3, 6}; // remove 1, 3, 6.
	int a[] = {0, 0, 1, 1, 2, 3, 3, 4}; // b remove 1, 3, 6.
	assertRequant(BSIZE(b), 3, r, 0, NULL, BSIZE(b), b, a);
}

void UtilTest::testRemoveAllButOne() {
	int r[] = {0, 1, 2, 4, 5, 6, 7}; // remove 0, 1, 2, 4, 5, 6, 7.
	int a[] = {0, 0, 0, 0, 0, 0, 0, 0}; // b remove 0, 1, 2, 4, 5, 6, 7.
	assertRequant(BSIZE(b), 7, r, 0, NULL, BSIZE(b), b, a);
}


void UtilTest::testInsertFirst() {
	int i = 0; // insert 0.
	int a[] = {1, 2, 3, 4, 5, 6, 7, 8}; // b insert 0.
	assertRequant(BSIZE(b), 0, NULL, 1, &i, BSIZE(b), b, a);
}

void UtilTest::testInsertLast() {
	int i = 8; // insert 8.
	assertRequant(BSIZE(b), 0, NULL, 1, &i, BSIZE(b), b, b);
}

void UtilTest::testInsertN() {
	int i[] = {2, 3, 5, 8, 12}; // insert 2, 3, 5, 8, 12.
	int a[] = {0, 1, 4, 6, 7, 9, 10, 11}; // b insert 2, 3, 5, 8, 12.
	assertRequant(BSIZE(b), 0, NULL, 5, i, BSIZE(b), b, a);
}

void UtilTest::testInsertAllButOne() {
	int i[] = {0, 1, 2, 3, 4, 6, 7}; // insert 0, 1, 2, 3, 4, 6, 7.
	int b1 = 0, a = 5; // {0} insert 0, 1, 2, 3, 4, 6, 7.
	assertRequant(1, 0, NULL, 7, i, 1, &b1, &a);
}


void UtilTest::testRearrange1() {
	int r = 2, i = 5; // remove 2 insert 5.
	int a[] = {0, 1, 1, 2, 3, 4, 6, 7}; // b remove 2 insert 5.
	assertRequant(BSIZE(b), 1, &r, 1, &i, BSIZE(b), b, a);
}

void UtilTest::testRearrangeN() {
	int r[] = {1, 3, 6}; // remove 1, 3, 6.
	int i[] = {0, 1, 5, 6, 8}; // insert 0, 1, 5, 6, 8.
	int a[] = {2, 2, 3, 3, 4, 7, 7, 9}; // b remove 1, 3, 6 insert 0, 1, 5, 6, 8.
	assertRequant(BSIZE(b), 3, r, 5, i, BSIZE(b), b, a);
}

void UtilTest::testRearrangeAllButOne() {
	int r[] = {0, 1, 2, 4, 5, 6, 7}; // remove 0, 1, 2, 4, 5, 6, 7.
	int i[] = {0, 1, 2, 3, 4, 6}; // insert 0, 1, 2, 3, 4, 6.
	int a[] = {5, 5, 5, 5, 5, 5, 5, 5}; // b remove 0, 1, 2, 4, 5, 6, 7 insert 0, 1, 2, 3, 4, 6.
	assertRequant(BSIZE(b), 7, r, 6, i, BSIZE(b), b, a);
}
