/*
Copyright (c) 2007 Johan Sarge

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#ifndef FIXER_BUFFER_HPP
#define FIXER_BUFFER_HPP

#include <cstring>

namespace buffer {
	
	template<typename E>
	int remove(int nBefore, E *buffer, int removed) {
		return remove(nBefore, buffer, 1, &removed);
	}
	
	// The buffer can be described by the regex K?(RK)*R?, where K is a
	// sequence of one or more Kept (not removed) elements and R is a
	// sequence of one or more Removed elements.
	template<typename E>
	int remove(int nBefore, E *buffer, int nRemoved, const int *removed) {
		if (nRemoved == 0) // Handle R-less cases.
			return nBefore;
		
		int
			nAfter = nBefore - nRemoved,
			srcIndexR = 0,
			srcIndexB = removed[0],
			trgIndex = removed[0], // Consume any initial K.
			endIndex,
			length;
		
		while (trgIndex < nAfter) { // Consume any and all RK pairs.
			do {
				srcIndexR++;
				srcIndexB++;
			} while (srcIndexR < nRemoved && srcIndexB == removed[srcIndexR]); // Consume R.
			
			endIndex = (srcIndexR == nRemoved) ? nBefore : removed[srcIndexR]; // Consume K.
			length = endIndex - srcIndexB;
			
			std::memmove(buffer + trgIndex, buffer + srcIndexB, length * sizeof (E));
			
			srcIndexB = endIndex;
			trgIndex += length;
		}
		
		// Ignore any final R.
		
		return nAfter;
	}
	
	template<typename E>
	int insert(int nBefore, E *buffer, int inserted, const E &insertedValue) {
		return insert(nBefore, buffer, 1, &inserted, &insertedValue);
	}
	
	// The buffer can be described by the regex K?(IK)*I?, where K is a
	// sequence of one or more places for Kept (not inserted) elements and I is a
	// sequence of one or more places for Inserted elements.
	template<typename E>
	int insert(int nBefore, E *buffer, int nInserted, const int *inserted, const E *insertedValues) {
		int nAfter = nBefore + nInserted;
		int
			srcIndexI = nInserted,
			srcIndexB = nBefore,
			trgIndex = nAfter,
			endIndex = nAfter,
			length;
		
		if (nInserted > 0 && nAfter - 1 == inserted[nInserted - 1]) { // Consume any final I.
			do {
				srcIndexI--;
				trgIndex--;
			} while (srcIndexI > 0 && trgIndex - 1 == inserted[srcIndexI - 1]);
			
			length = endIndex - trgIndex;
			
			std::memcpy(buffer + trgIndex, insertedValues + srcIndexI, length * sizeof (E));
			
			endIndex = trgIndex;
		}
		
		while (srcIndexI > 0) { // Consume any and all IK pairs.
			trgIndex = inserted[srcIndexI - 1] + 1; // Consume K.
			length = endIndex - trgIndex;
			srcIndexB -= length;
			
			std::memmove(buffer + trgIndex, buffer + srcIndexB, length * sizeof (E));
			
			endIndex = trgIndex;
			
			do {
				srcIndexI--;
				trgIndex--;
			} while (srcIndexI > 0 && trgIndex - 1 == inserted[srcIndexI - 1]); // Consume I.
			
			length = endIndex - trgIndex;
			
			std::memcpy(buffer + trgIndex, insertedValues + srcIndexI, length * sizeof (E));
			
			endIndex = trgIndex;
		}
		
		// Ignore any initial K.
		
		return nAfter;
	}
	
	template<typename E>
	int rearrange(int nBefore, E *buffer, int removed, int inserted, const E &insertedValue) {
		return insert(remove(nBefore, buffer, removed), buffer, inserted, insertedValue);
	}
	
	template<typename E>
	int rearrange(
	int nBefore, E *buffer, int nRemoved, const int *removed,
	int nInserted, const int *inserted, const E *insertedValues) {
		return
			insert(remove(nBefore, buffer, nRemoved, removed), buffer, nInserted, inserted, insertedValues);
	}
}

#endif
