/*
Copyright (c) 2007 Johan Sarge

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#include "fixerutils.hpp"

#include <cstring>
#include <iomanip>
#include <iostream>

using namespace std;

void printHex(int nBytes, const char *bytes) {
	cout << hex << setfill('0');
	for (int n = 0; n < nBytes; n++) {
		cout << setw(2) << ((int) bytes[n] + 256) % 256;
		
		if ((n+1) % 16 == 0)
			cout << "  " << n + 1 << endl;
		else if ((n+1) % 4 == 0)
			cout << " ";
	}
	cout << endl << dec;
}

bool endsWith(const char *str, const char *suffix) {
	int slen = strlen(str), suffixlen = strlen(suffix);
	return slen >= suffixlen && !strncmp(str + (slen - suffixlen), suffix, suffixlen);
}

// In ".fxp" preset files, both integers and IEEE floating point numbers are
// stored in big-endian byte order. Since x86 processors are little-endian,
// these functions are needed on that platform.
int reverseByteOrder(int i) {
	// NOTE: If we use signed int we get sign-preserving shifts, which is bad in this case.
	unsigned int ui = (unsigned int) i;
	return ui << 24 | (ui & 0xff00) << 8 | (ui & 0xff0000) >> 8 | ui >> 24;
}

int reverseByteOrderF(float f) {
	return reverseByteOrder(*((int *) &f));
}

float reverseByteOrderF(int i) {
	i = reverseByteOrder(i);
	return *((float *) &i);
}

float requantize(
	int nBefore, float before, int nRemoved, const int *removed, int nInserted, const int *inserted)
{
	// Scale float value in [0,1] range to step number. Ensure that iAfter < nBefore.
	int iAfter = (int) (0.999f * (before * nBefore));
	
	// --- Handle removed steps ---
	// Decrement step number once for each removed step <= the current step.
	for (int i = nRemoved - 1; i >= 0; i--)
		if (removed[i] <= iAfter)
			iAfter--;
	
	if (iAfter == -1) // Handle the case where all steps <= the current step were removed.
		iAfter = 0; // Select the lowest remaining step.
	
	// --- Handle inserted steps ---
	// Increment step number once for each inserted step <= the current step.
	for (int i = 0; i < nInserted && inserted[i] <= iAfter; i++)
		iAfter++;
	
	// Scale step number to float value in [0,1] range.
	return (iAfter + 0.5f) / (nBefore + nInserted - nRemoved);
}

void requantize(
	int nBefore, int *before, int nRemoved, const int *removed, int nInserted, const int *inserted)
{
	*before = reverseByteOrderF(requantize(
		nBefore, reverseByteOrderF(*before), nRemoved, removed, nInserted, inserted));
}
