/*
Copyright (c) 2006 Johan Sarge

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit
persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#include <algorithm>
#include <cstdio>
#include <cstring>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <vector>
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "fixerutils.hpp"
#include "ConverterRegistry.hpp"

using namespace std;

// ---<<< Program data (configuration, data buffer, converters) >>>---
#define DEF_ERRORLEVEL 2
#define DEF_VERBOSITY 0
#define DEF_TARGET -1

bool help = false, conversionList = false, overwrite = false, dowrite = true, recursive = false;
int errorlevel = DEF_ERRORLEVEL, verbosity = DEF_VERBOSITY, targetID = DEF_TARGET;
const char *outputPrefix = NULL, *dataFileSuffixes[] = {".fxp", ".FXP", NULL};

char fxpdata[2048];

vector<Converter *> converters, chain;


// ---<<< Format recognition and conversion functions >>>---
bool isValidTargetFormat(int target) {
	for (vector<Converter *>::const_iterator i = converters.begin(); i != converters.end(); i++)
		if (target == (*i)->getTargetID())
			return true;
	
	return false;
}

bool isCycleFree(bool checkIncreasingID, const vector<Converter *> &visited, Converter *next) {
	if (checkIncreasingID)
		return next->getSourceID() < next->getTargetID();
	
	for (vector<Converter *>::const_iterator i = visited.begin(); i != visited.end(); i++)
		if (next->getTargetID() == (*i)->getSourceID())
			return false;
	
	return true;
}

bool findConverterChain(
	int source, int target, vector<Converter *> &visited, int depth, int maxDepth)
{
	if (source == target)
		return true;
	
	if (depth == maxDepth)
		return target == -1;
	
	for (vector<Converter *>::const_iterator i = converters.begin(); i != converters.end(); i++) {
		if ((*i)->getSourceID() == source && isCycleFree(target == -1, visited, *i)) {
			visited.push_back(*i);
			
			if (findConverterChain((*i)->getTargetID(), target, visited, depth + 1, maxDepth))
				return true;
			
			visited.pop_back();
		}
	}
	
	return target == -1;
}

bool makeConverterChain(int source, int target, vector<Converter *> &output) {
	if (target > 0) {
		for (int maxDepth = 0; maxDepth <= converters.size(); maxDepth++)
			if (findConverterChain(source, target, output, 0, maxDepth))
				return true;
		
		return false;
	}
	
	return findConverterChain(source, -1, output, 0, converters.size());
}


// ---<<< File processing functions >>>---
const char *makeOutputPath(int outsize, char *out, const char *prefix, const char *in) {
	int prefixlen = strlen(prefix), inlen = strlen(in);
	
	if (overwrite) {
		if (inlen >= outsize)
			return NULL;
		
		memcpy(out, in, inlen + 1); // Include zero-termination.
	}
	else {
		if (prefixlen + inlen >= outsize)
			return NULL;
		
		char *slashPos = strrchr(in, '/'), *bslashPos = strrchr(in, '\\');
		int nameIndex = 0;
		
		if (slashPos > bslashPos)
			nameIndex = (slashPos - in) + 1;
		else if (bslashPos > slashPos)
			nameIndex = (bslashPos - in) + 1;
		// else (slashPos == bslashPos == NULL). No directory names in input path.
		
		memcpy(out, in, nameIndex);
		memcpy(out + nameIndex, prefix, prefixlen);
		memcpy(out + nameIndex + prefixlen, in + nameIndex, inlen - nameIndex + 1);
	}
	
	return out;
}

const char *pathConcat(int outsize, char *out, const char *in1, const char *in2) {
	int in1len = strlen(in1), in2len = strlen(in2);
	
	if (in1len + 1 + in2len >= outsize)
		return NULL;
	
	memcpy(out, in1, in1len);
	out[in1len] = '\\';
	memcpy(out + in1len + 1, in2, in2len + 1); // Include zero-termination.
	
	return out;
}

bool isDataFile(const char *path) {
	for (const char **suffix = dataFileSuffixes; *suffix != NULL; suffix++)
		if (endsWith(path, *suffix))
			return true;
	
	return false;
}

bool isDirectory(const char *path) {
	int fattr = GetFileAttributes(path);
	return fattr != 0xffffffff && fattr & FILE_ATTRIBUTE_DIRECTORY;
}

bool processDataFile(const char *filepath) {
	if (verbosity > 1)
		cout << "Loading data file: " << filepath << endl;
	
	ifstream in(filepath, ios::in | ios::binary);
	
	if (!in.good()) {
		cerr << "Error 2: Failed to open input file: " << filepath << endl;
		return errorlevel > 2;
	}
	
	int nBytes = 0;
	while (nBytes < sizeof fxpdata && in.get(fxpdata[nBytes]))
		nBytes++;
	
	in.close();
	
	if (verbosity > 2)
		printHex(nBytes, fxpdata);
	
	int sourceID = -1;
	const char *sourceText;
	
	for (vector<Converter *>::iterator i = converters.begin(); i != converters.end(); i++)
		if ((*i)->isConvertable(nBytes, fxpdata)) {
			sourceID = (*i)->getSourceID();
			sourceText = (*i)->getSourceDescription();
			break;
		}
	
	if (sourceID == -1) {
		cerr << "Error 1: Unrecognized source format: " << filepath << endl;
		return errorlevel > 1;
	}
	
	chain.clear();
	if (!makeConverterChain(sourceID, targetID, chain)) {
		cerr << "Error 1: Couldn't make converter chain: " << filepath << endl;
		return errorlevel > 1;
	}
	
	if (chain.size() == 0) // Nothing to be done with this file.
		return true;
	
	if (verbosity == 1)
		cout << "Converting data file: " << filepath << endl;
	
	// Perform the conversion(s).
	for (vector<Converter *>::iterator i = chain.begin(); i != chain.end(); i++) {
		if (verbosity > 1)
			cout << (*i)->getSourceDescription() << " -> " << (*i)->getTargetDescription() << endl;
		
		if (!(*i)->isConvertable(nBytes, fxpdata)) {
			cerr << "Error 2: Converter input does not match source format: " << filepath << endl;
			return errorlevel > 2;
		}
		
		nBytes = (*i)->convert(nBytes, sizeof fxpdata, fxpdata);
		
		if (nBytes <= 0) {
			cerr << "Error 2: Conversion failed: " << filepath << endl;
			return errorlevel > 2;
		}
		
		if (verbosity > 3)
			printHex(nBytes, fxpdata);
	}
	
	if (verbosity == 3)
		printHex(nBytes, fxpdata);
	
	const char *prefix = outputPrefix;
	char defaultPrefix[16];
	if (prefix == NULL) {
		sprintf(defaultPrefix, "%d_", chain.back()->getTargetID() % 1000000);
		prefix = defaultPrefix;
	}
	
	char outpath[MAX_PATH];
	if (!makeOutputPath(MAX_PATH, outpath, prefix, filepath)) {
		cerr << "Error 2: Couldn't make output path: " << filepath << endl;
		return errorlevel > 2;
	}
	
	if (dowrite) {
		ofstream out(outpath, ios::out | ios::binary);
		
		if (!out.good()) {
			cerr << "Error 2: Failed to open output file: " << outpath << endl;
			return errorlevel > 2;
		}
		
		out.write(fxpdata, nBytes);
		
		out.close();
	}
	
	return true;
}

bool processDirectory(const char *filepath) {
	if (verbosity > 0)
		cout << "Searching directory: " << filepath << endl;
	
	char nextpath[MAX_PATH];
	
	if (!pathConcat(MAX_PATH, nextpath, filepath, "*")) {
		cerr << "Error 2: Couldn't make directory path: " << filepath << endl;
		return errorlevel > 2;
	}
	
	WIN32_FIND_DATA fdata;
	HANDLE fhandle = FindFirstFile(nextpath, &fdata);
	
	if (fhandle == INVALID_HANDLE_VALUE) {
		if (verbosity > 0)
			cout << "No files found." << endl;
		return true;
	}
	
	do {
		if (pathConcat(MAX_PATH, nextpath, filepath, fdata.cFileName)) {
			//cout << "Found file: " << nextpath << endl; // DEBUG: 
			
			if (isDataFile(nextpath)) {
				if (!processDataFile(nextpath))
					goto handle_failure;
			}
			else if (recursive && fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
			         strcmp(fdata.cFileName, ".") &&  strcmp(fdata.cFileName, "..")) {
				if (!processDirectory(nextpath))
					goto handle_failure;
			}
		}
		else {
			cerr << "Error 2: Couldn't make file path: " << fdata.cFileName << endl;
			if (errorlevel < 3)
				goto handle_failure;
		}
	} while (FindNextFile(fhandle, &fdata));
	
	FindClose(fhandle);
	return true;
	
	handle_failure:
	FindClose(fhandle);
	return false;
}

bool processFile(const char *filepath) {
	if (isDataFile(filepath))
		return processDataFile(filepath);
	else if (isDirectory(filepath))
		return processDirectory(filepath);
	else
		return true;
}


// ---<<< UI functions >>>---
void printHelp() {
	cout << "Usage: presetfixer [OPTION]... [FILE]..." << endl;
	cout << endl;
	cout << "-?            Print this help message and exit." << endl;
	cout << endl;
	cout << "-eERRORLEVEL  Abort on errors with severity >= the specified level." << endl;
	cout << endl;
	cout << "-l            Print a list of available converters and exit." << endl;
	cout << endl;
	cout << "-n            Don't write updated presets back to file. Useful for testing." << endl;
	cout << endl;
	cout << "-o            Overwrite old preset files with successfully updated presets." << endl;
	cout << endl;
	cout << "-pPREFIX      Use this prefix for updated presets instead of the default." << endl;
	cout << endl;
	cout << "-r            Scan subdirectories recursively and update presets within them." << endl;
	cout << endl;
	cout << "-tFORMAT      Convert presets to the specified format." << endl;
	cout << endl;
	cout << "-v            Be verbose. Be even more verbose if this option is repeated." << endl;
	cout << endl;
	cout << "FILE          A preset to convert or a directory to scan for presets." << endl;
}

void printConversions() {
	for (vector<Converter *>::iterator i = converters.begin(); i != converters.end(); i++)
		cout <<
			(*i)->getSourceDescription() << " (" << (*i)->getSourceID() << ") -> " <<
			(*i)->getTargetDescription() << " (" << (*i)->getTargetID() << ")" << endl;
}

const char *getOptionArgument(const char *str, int &index) {
	const char *tmp = str + index;
	index += strlen(tmp);
	return tmp;
}

int parseOptions(int argc, const char **argv) {
	int firstSource = 1, argIndex = 1;
	
	while (firstSource < argc && argv[firstSource][0] == '-') {
		char optionSymbol = argv[firstSource][argIndex++];
		const char *optionArg;
		
		switch (optionSymbol) {
			case '\0': // No more options in this argument.
			firstSource++;
			argIndex = 1;
			break;
			
			case '?': case 'h': help = true;
			break;
			
			case 'e':
			optionArg = getOptionArgument(argv[firstSource], argIndex);
			if (sscanf(optionArg, "%d", &errorlevel) != 1) {
				cerr << "Error 3: Invalid error level '" << optionArg << "'." << endl;
				if (errorlevel < 4)
					return 0;
				errorlevel = DEF_ERRORLEVEL;
			}
			break;
			
			case 'l': conversionList = true;
			break;
			
			case 'n': dowrite = false;
			break;
			
			case 'o': overwrite = true;
			break;
			
			case 'p':
			outputPrefix = getOptionArgument(argv[firstSource], argIndex);
			if (*outputPrefix == '\0') {
				cerr << "Error 3: '-p' option requires nonempty parameter." << endl;
				if (errorlevel < 4)
					return 0;
				outputPrefix = NULL;
			}
			break;
			
			case 'r': recursive = true;
			break;
			
			case 't':
			optionArg = getOptionArgument(argv[firstSource], argIndex);
			if (sscanf(optionArg, "%d", &targetID) != 1 || !isValidTargetFormat(targetID)) {
				cerr << "Error 3: Invalid target format '" << optionArg << "'." << endl;
				if (errorlevel < 4)
					return 0;
				targetID = DEF_TARGET;
			}
			break;
			
			case 'v': verbosity++;
			break;
			
			default:
			cerr << "Error 3: Unrecognized option '-" << optionSymbol << "'." << endl;
			if (errorlevel < 4)
				return 0;
		}
	}
	
	return firstSource;
}

bool converterLess(Converter *c1, Converter *c2) {
	return
		c1->getSourceID() < c2->getSourceID() ||
		c1->getSourceID() == c2->getSourceID() && c1->getTargetID() < c2->getTargetID();
}

int main(int argc, const char **argv) {
	
	ConverterRegistry::getConverters(converters);
	sort(converters.begin(), converters.end(), &converterLess);
	
	int sourceIndex = parseOptions(argc, argv);
	if (!sourceIndex)
		return 1;
	
	if (help)
		printHelp();
	
	if (conversionList)
		printConversions();
	
	if (help | conversionList)
		return 0;
	
	if (overwrite & outputPrefix != NULL) {
		cerr << "Error 3: Prefix and overwrite options combined." << endl;
		if (errorlevel < 4)
			return 2;
	}
	
	if (sourceIndex >= argc) {
		if (!processFile("."))
			return 3;
	}
	else {
		while (sourceIndex < argc)
			if (!processFile(argv[sourceIndex++]))
				return 3;
	}
	
	return 0;
}
